@@ -1,4 +1,6 @@
 import re
+import sys
+from unicodedata import category
 
 
 whitespace = {' ', '\t', '\n', '\x0b', '\x0c', '\r'}
@@ -6,9 +8,18 @@ unicode_whitespace = {'\t', '\n', '\x0b', '\x0c', '\r', '\x1c', '\x1d', '\x1e',
         '\x1f', ' ', '\x85', '\xa0', '\u1680', '\u2000', '\u2001', '\u2002',
         '\u2003', '\u2004', '\u2005', '\u2006', '\u2007', '\u2008', '\u2009',
         '\u200a', '\u2028', '\u2029', '\u202f', '\u205f', '\u3000'}
-punctuation = {'!', '"', '#', '$', '%', '&', '\'', '(', ')', '*', '+', ',',
-               '-', '.', '/', ':', ';', '<', '=', '>', '?', '@', '[', '\\',
-               ']', '^', '_', '`', '{', '|', '}', '~'}
+
+# punctuation: _ASCII and Unicode punctuation characters_ as defined at
+# <https://spec.commonmark.org/0.30/#ascii-punctuation-character> and
+# <https://spec.commonmark.org/0.30/#unicode-punctuation-character>
+unicode_chrs = (chr(i) for i in range(sys.maxunicode + 1))
+punctuation = set.union(
+        {'!', '"', '#', '$', '%', '&', '\'', '(', ')', '*', '+', ',',
+        '-', '.', '/', ':', ';', '<', '=', '>', '?', '@', '[', '\\',
+        ']', '^', '_', '`', '{', '|', '}', '~'},
+        {c for c in unicode_chrs if category(c).startswith("P")},
+)
+
 code_pattern = re.compile(r"(?<!\\|`)(?:\\\\)*(`+)(?!`)(.+?)(?<!`)\1(?!`)", re.DOTALL)
 
 
@@ -27,7 +38,7 @@ def find_core_tokens(string, root):
     while i < len(string):
         if code_match is not None and i == code_match.start():
             if in_delimiter_run is not None:
-                delimiters.append(Delimiter(start, i if not escaped else i-1, string))
+                delimiters.append(Delimiter(start, i if not escaped else i - 1, string))
                 in_delimiter_run = None
                 escaped = False
             _code_matches.append(code_match)
@@ -40,7 +51,7 @@ def find_core_tokens(string, root):
             i += 1
             continue
         if in_delimiter_run is not None and (c != in_delimiter_run or escaped):
-            delimiters.append(Delimiter(start, i if not escaped else i-1, string))
+            delimiters.append(Delimiter(start, i if not escaped else i - 1, string))
             in_delimiter_run = None
         if in_delimiter_run is None and (c == '*' or c == '_') and not escaped:
             in_delimiter_run = c
@@ -48,9 +59,9 @@ def find_core_tokens(string, root):
         if not escaped:
             if c == '[':
                 if not in_image:
-                    delimiters.append(Delimiter(i, i+1, string))
+                    delimiters.append(Delimiter(i, i + 1, string))
                 else:
-                    delimiters.append(Delimiter(i-1, i+1, string))
+                    delimiters.append(Delimiter(i - 1, i + 1, string))
                     in_image = False
             elif c == '!':
                 in_image = True
@@ -110,11 +121,12 @@ def process_emphasis(string, stack_bottom, delimiters, matches):
             n = 2 if closer.number >= 2 and opener.number >= 2 else 1
             start = opener.end - n
             end = closer.start + n
-            match = MatchObj(start, end, (start+n, end-n, string[start+n:end-n]))
+            match = MatchObj(start, end, (start + n, end - n, string[start + n:end - n]))
             match.type = 'Strong' if n == 2 else 'Emphasis'
+            match.delimiter = string[start]
             matches.append(match)
             # remove all delimiters in between
-            del delimiters[open_pos+1:curr_pos]
+            del delimiters[open_pos + 1:curr_pos]
             curr_pos -= curr_pos - open_pos - 1
             # remove appropriate number of chars from delimiters
             if not opener.remove(n, left=False):
@@ -148,7 +160,7 @@ def match_link_image(string, offset, delimiter, root=None):
     # inline link
     if follows(string, offset, '('):
         # link destination
-        match_info = match_link_dest(string, offset+1)
+        match_info = match_link_dest(string, offset + 1)
         if match_info is not None:
             dest_start, dest_end, dest = match_info
             # link title
@@ -164,11 +176,13 @@ def match_link_image(string, offset, delimiter, root=None):
                                       (dest_start, dest_end, dest),
                                       (title_start, title_end, title))
                     match.type = 'Link' if not image else 'Image'
+                    match.dest_type = "angle_uri" if dest_start < dest_end and string[dest_start] == "<" else "uri"
+                    match.title_delimiter = string[title_start] if title_start < title_end else None
                     return match
     # footnote link
     if follows(string, offset, '['):
-        # full footnote link
-        result = match_link_label(string, offset+1, root)
+        # full footnote link: [label][dest]
+        result = match_link_label(string, offset + 1, root)
         if result:
             match_info, (dest, title) = result
             end = match_info[1]
@@ -177,11 +191,13 @@ def match_link_image(string, offset, delimiter, root=None):
                               (-1, -1, dest),
                               (-1, -1, title))
             match.type = 'Link' if not image else 'Image'
+            match.label = match_info[2]
+            match.dest_type = "full"
             return match
-        ref = is_link_label(text, root)
+        ref = get_link_label(text, root)
         if ref:
-            # compact footnote link
-            if follows(string, offset+1, ']'):
+            # compact (collapsed) footnote link: [dest][]
+            if follows(string, offset + 1, ']'):
                 dest, title = ref
                 end = offset + 3
                 match = MatchObj(start, end,
@@ -189,10 +205,11 @@ def match_link_image(string, offset, delimiter, root=None):
                                   (-1, -1, dest),
                                   (-1, -1, title))
                 match.type = 'Link' if not image else 'Image'
+                match.dest_type = "collapsed"
                 return match
         return None
-    # shortcut footnote link
-    ref = is_link_label(text, root)
+    # shortcut footnote link: [dest]
+    ref = get_link_label(text, root)
     if ref:
         dest, title = ref
         end = offset + 1
@@ -201,23 +218,24 @@ def match_link_image(string, offset, delimiter, root=None):
                           (-1, -1, dest),
                           (-1, -1, title))
         match.type = 'Link' if not image else 'Image'
+        match.dest_type = "shortcut"
         return match
     return None
 
 
 def match_link_dest(string, offset):
-    offset = shift_whitespace(string, offset+1)
+    offset = shift_whitespace(string, offset + 1)
     if offset == len(string):
         return None
     if string[offset] == '<':
         escaped = False
-        for i, c in enumerate(string[offset+1:], start=offset+1):
+        for i, c in enumerate(string[offset + 1:], start=offset + 1):
             if c == '\\' and not escaped:
                 escaped = True
-            elif c == ' ' or c == '\n' or (c == '<' and not escaped):
+            elif c == '\n' or (c == '<' and not escaped):
                 return None
             elif c == '>' and not escaped:
-                return offset, i+1, string[offset+1:i]
+                return offset, i + 1, string[offset + 1:i]
             elif escaped:
                 escaped = False
         return None
@@ -258,11 +276,11 @@ def match_link_title(string, offset):
     else:
         return None
     escaped = False
-    for i, c in enumerate(string[offset+1:], start=offset+1):
+    for i, c in enumerate(string[offset + 1:], start=offset + 1):
         if c == '\\' and not escaped:
             escaped = True
         elif c == closing and not escaped:
-            return offset, i+1, string[offset+1:i]
+            return offset, i + 1, string[offset + 1:i]
         elif escaped:
             escaped = False
     return None
@@ -282,8 +300,8 @@ def match_link_label(string, offset, root=None):
                 return None
         elif c == ']' and not escaped:
             end = i
-            label = string[start+1:end]
-            match_info = start, end+1, label
+            label = string[start + 1:end]
+            match_info = start, end + 1, label
             if label.strip() != '':
                 ref = root.footnotes.get(normalize_label(label), None)
                 if ref is not None:
@@ -295,7 +313,13 @@ def match_link_label(string, offset, root=None):
     return None
 
 
-def is_link_label(text, root):
+def get_link_label(text, root):
+    """
+    Normalize and look up `text` among the footnotes.
+    Returns (destination, title) if successful, otherwise None.
+    """
+    if not root:
+        return None
     escaped = False
     for c in text:
         if c == '\\' and not escaped:
@@ -305,8 +329,6 @@ def is_link_label(text, root):
         elif escaped:
             escaped = False
     if text.strip() != '':
-        if not root:
-            return True
         return root.footnotes.get(normalize_label(text), None)
     return None
 
@@ -326,7 +348,7 @@ def matching_opener(curr_pos, delimiters, bottom):
     if curr_pos > 0:
         curr_delimiter = delimiters[curr_pos]
         index = curr_pos - 1
-        for delimiter in delimiters[curr_pos-1:bottom:-1]:
+        for delimiter in delimiters[curr_pos - 1:bottom:-1]:
             if (hasattr(delimiter, 'open')
                     and delimiter.open
                     and delimiter.closed_by(curr_delimiter)):
@@ -368,7 +390,7 @@ def is_right_delimiter(start, end, string):
 
 
 def preceded_by(start, string, charset):
-    preceding_char = string[start-1] if start > 0 else ' '
+    preceding_char = string[start - 1] if start > 0 else ' '
     return preceding_char in charset
 
 
@@ -382,7 +404,7 @@ def is_control_char(char):
 
 
 def follows(string, index, char):
-    return index + 1 < len(string) and string[index+1] == char
+    return index + 1 < len(string) and string[index + 1] == char
 
 
 def shift_whitespace(string, index):
@@ -423,10 +445,16 @@ class Delimiter:
         return True
 
     def closed_by(self, other):
-        return not (self.type[0] != other.type[0]
-                    or (self.open and self.close or other.open and other.close)
-                    and (self.number + other.number) % 3 == 0)
-
+        if self.type[0] != other.type[0]:
+            return False
+        if self.open and self.close or other.open and other.close:
+            # if either of the delimiters can both open and close emphasis, then additional
+            # restrictions apply: the sum of the lengths of the delimiter runs
+            # containing the opening and closing delimiters must not be a multiple of 3
+            # unless both lengths are multiples of 3.
+            return ((self.number + other.number) % 3 != 0
+                    or (self.number % 3 == 0 and other.number % 3 == 0))
+        return True
 
     def __repr__(self):
         if not self.type.startswith(('*', '_')):
@@ -443,17 +471,17 @@ class MatchObj:
     def start(self, n=0):
         if n == 0:
             return self._start
-        return self.fields[n-1][0]
+        return self.fields[n - 1][0]
 
     def end(self, n=0):
         if n == 0:
             return self._end
-        return self.fields[n-1][1]
+        return self.fields[n - 1][1]
 
     def group(self, n=0):
         if n == 0:
             return ''.join([field[2] for field in self.fields])
-        return self.fields[n-1][2]
+        return self.fields[n - 1][2]
 
     def __repr__(self):
         return '<MatchObj fields={} start={} end={}>'.format(self.fields, self._start, self._end)
